;
; Garg Demo!
; by Bart Trzynadlowski
; December 20, 2005 and January 7, 2006
;

;
; demo.asm
;
; Let's you move a Garg around the screen. The sprites were ripped from 
; Commander Keen: Invasion of the Vorticons by id Software.
;


;******************************************************************************
; Constants
;******************************************************************************

;
; Variables
;
; Addresses for variables stored in RAM.
;

VBL = $F001				; VBL counter

PLAYER_X = $F002		; X position of player
PLAYER_Y = $F003		; Y position
PLAYER_VX = $F004		; velocity X
PLAYER_VY = $F005		; velocity Y
PLAYER_JUMP = $F006		; incremented each frame jump button is held down for

PLAYER_X_PREV = $F007	; X position during previous update cycle
PLAYER_Y_PREV = $F008	; X position during previous update cycle

ANIM_PTR = $F00A		; base of 2-byte pointer to animation sequence

;
; Constants
;

SPRITE_WIDTH = 23		; width of sprite in pixels
SPRITE_HEIGHT = 31		; height
GRAVITY = 1				; acceleration due to gravity
TERMINAL_VELOCITY = 9	; maximum vertical speed
WALK_SPEED = 1			; walking speed (MUST be < SPRITE_WIDTH/2)
MAX_JUMP = 0			; ??? maximum velocity due to jump


;******************************************************************************
; Vector Table
;******************************************************************************

;
; ResetVector:
;
; Execution begins here when the Z80 powers up.
;

.ORG $0
ResetVector:
    di              ; disable interrupts
    ld  sp,$F400    ; set SP to the top of the RAM area
	jp	Start
	
;
; InterruptVector:
;
; Do nothing.
;

.ORG $38
InterruptVector:
	reti
	
;
; NMIVector:
;
; Triggered at the start of the VBlank period. Increments VBL.
;
.ORG $66
NMIVector:
	push	af
	ld		a,(VBL)
	inc		a
	ld		(VBL),a
	pop		af	
	reti

;******************************************************************************
; Main Program
;******************************************************************************

;
; Start:
;
; Program entry point. 
;

Start:
	;
	; Clear screen and enable interrupts
	;
	ld		hl,$9000	; control reg	
	ld		(hl),4		; settings mode
	ld		a,0
	ld		($8000),a	; disable interrupts and turn screen off
	call	ClearScreen	
	ld		(hl),4		
	ld		a,3
	ld		($8000),a	; turn display and interrupts on	
	
	;
	; Initialize player
	;
	ld		a,0
	ld		(PLAYER_X),a
	ld		(PLAYER_Y),a
	ld		(PLAYER_VX),a
	ld		(PLAYER_VY),a
	ld		(PLAYER_JUMP),a
	ld		hl,anim_idle	; get address of idle animation
	ld		(ANIM_PTR),hl	; store as current animation pointer
	
	;
	; Main Loop:
	;
	; - Read inputs
	; - Update player velocities
	; - Apply acceleration
	; - Save old coordinates
	; - Update player coordinates
	; - Perform out-of-bounds checks and clamp appropriately
	; - Wait until VBL
	; - Erase sprite
	; - Draw sprite at updated coordinates
	;	
MainLoop:

	;
	; Read joypad
	;
	ld		hl,$9000	; control reg
	ld		de,$A000	; joypad status
	ld		(hl),8		; mux = 1
	ld		a,(de)		; A = --UDLRBC
	ld		b,a			; save copy in B
	
	;
	; Update player velocities
	;
	ld		hl,anim_idle	; default animation
	ld		(ANIM_PTR),hl
	and		$08			; if LEFT is pressed (0), move left
	ld		a,0
	jr		nz,no_left
	ld		hl,anim_walk_left	; queue animation...
	ld		(ANIM_PTR),hl
	ld		a,-WALK_SPEED
no_left:
	ld		(PLAYER_VX),a
	ld		a,b			; get button data
	and		$04			; if RIGHT is pressed, move right
	jr		nz,no_right
	ld		hl,anim_walk_right	; queue animation...
	ld		(ANIM_PTR),hl
	ld		a,WALK_SPEED
	ld		(PLAYER_VX),a
no_right:
	
	;
	; TO-DO: Implement jumping...
	;
	
	;
	; Apply acceleration (jumping and gravity)
	;
	ld		a,(PLAYER_VY)
	add		a,GRAVITY
	cp		TERMINAL_VELOCITY	; check for terminal velocity
	jr		nc,dont_clamp_vy
	ld		a,TERMINAL_VELOCITY	; if reached, clamp the velocity to that
dont_clamp_vy:
	ld		(PLAYER_VY),a
	
	;
	; Save old coordinates
	;
	ld		a,(PLAYER_X)
	ld		(PLAYER_X_PREV),a
	ld		a,(PLAYER_Y)
	ld		(PLAYER_Y_PREV),a
	
	;
	; Update player coordinates (ie., apply velocities!)
	;
	ld		a,(PLAYER_X)
	ld		b,a
	ld		a,(PLAYER_VX)
	add		a,b
	ld		(PLAYER_X),a
	ld		a,(PLAYER_Y)
	ld		b,a
	ld		a,(PLAYER_VY)
	add		a,b
	ld		(PLAYER_Y),a
	
	;
	; Perform out-of-bounds checks.
	;
	;
	; There is a trick here: We merely assume that the maximum X coordinate is
	; 255 for the purpose of this demo. Therefore, if the player X exceeds 
	; (256 - SPRITE_WIDTH/2), we assume that the player has wandered too far 
	; left and clamp to 0. Then, we check for the coordinate exceeding
	; (256 - SPRITE_WIDTH) and clamp there.
	;
	; This is all under the assumption that WALK_SPEED < SPRITE_WIDTH/2.
	;
	
	ld		a,(PLAYER_X)
	cp		256-(SPRITE_WIDTH/2)
	jr		c,no_clamp_left		; if carry, no need to clamp to X=0
	ld		a,0
	ld		(PLAYER_VX),a		; if clamping X, also kill velocity
no_clamp_left:	
	cp		(256-SPRITE_WIDTH)
	jr		c,no_clamp_right	; if carry, we are to the left of the test pt.
	ld		a,0
	ld		(PLAYER_VX),a		; kill velocity
	ld		a,256-SPRITE_WIDTH
no_clamp_right:
	ld		(PLAYER_X),a		; write back the clamped X coordinate
	
	ld		a,(PLAYER_Y)		; clamp Y coordinate now to bottom of screen
	cp		240-SPRITE_HEIGHT
	jr		c,no_clamp_bottom
	ld		a,0
	ld		(PLAYER_VY),a
	ld		a,240-SPRITE_HEIGHT
no_clamp_bottom:
	ld		(PLAYER_Y),a
	
	;
	; Wait until VBL
	;
	call	WaitVBL
	call	WaitVBL
	call	WaitVBL
	
	;
	; Erase sprite at old location and draw new one
	;
	ld		b,23
	ld		c,31
	ld		a,(PLAYER_X_PREV)
	ld		d,a
	ld		a,(PLAYER_Y_PREV)
	ld		e,a
	ld		hl,spr_blank
	call	DrawSprite
	
	;
	; Draw sprite at new location using current animation frame
	;
	ld		a,(VBL)		; use refresh rate / 8, and then multiply by 2 because	
	srl		a			; each pointer is 2 bytes. So only shift twice.
	srl		a
	and		$06			; only 4 animations...
	ld		e,a
	ld		d,0
	ld		hl,(ANIM_PTR)	; get base of animation
	add		hl,de			; add in the offset	
	ld		a,(hl)			; get low byte of sprite address
	ld		e,a
	inc		hl
	ld		a,(hl)
	ld		d,a
	push	de
	pop		hl				; transfer address of sprite to HL

	ld		a,(PLAYER_X)
	ld		d,a
	ld		a,(PLAYER_Y)
	ld		e,a
	call	DrawSprite

	jp		MainLoop

	
;******************************************************************************
; Video Functions
;******************************************************************************

;
; WaitVBL:
;
; Waits until the next VBlank period has started.
;
WaitVBL:
	push	af
	push	bc
	ld		a,(VBL)
	ld		b,a			; old VBL counter into B
WaitVBL_loop:
	ld		a,(VBL)
	cp		b
	jr		z,WaitVBL_loop
	pop		bc
	pop		af
	ret
	
;
; DrawSprite: 
;
; Draws a sprite on the screen.
;
; Input:
;   B  = X size.
;   C  = Y size.
;	D  = X position.	
;   E  = Y position.
;   HL = Sprite address.
;
; Saves AF, BC, DE, HL.
;
DrawSprite:
    push	af
    push	bc
    push	de
    push	hl
    
    push	hl
    push	bc
    ld		bc,$8000	; data register
    ld		hl,$9000	; command register    
    ld		(hl),2		; set X position high to 0
    ld		a,0
    ld		(bc),a
    ld		(hl),1		; set X position low
    ld		a,d
    ld		(bc),a
	pop		bc
    pop		hl
    
yloop:
	ld		a,3			; set Y position
	ld		($9000),a
	ld		a,e
	ld		($8000),a
	
	ld		a,0			; data write mode
	ld		($9000),a
	push	bc			; save X size

xloop:
	ld		a,(hl)		; fetch pixel
	inc		hl
	ld		($8000),a	; write to framebuffer
	dec		b			; decrement X count and loop
	ld		a,b
	cp		0
	jr		nz,xloop

	pop		bc
	inc		e			; next line in framebuffer
	dec		c			; decrement Y count and loop
	ld		a,c
	cp		0
	jr		nz,yloop
	
	pop		hl
	pop		de
	pop		bc
	pop		af
	ret
;
; ClearScreen:
;
; Clears the screen to black. Saves BC, HL, DE and destroys everything else.
;
ClearScreen:
	push	bc
	push	hl
	push	de
	
	ld		a,1
	ld		($9000),a	; set address X low
	ld		a,0
	ld		($8000),a
	ld		a,2			; set address X high
	ld		($9000),a
	ld		a,0
	ld		($8000),a
	
	ld		b,239	; 240 lines, start clearing at last one
	ld		de,1
ClearScreen_yloop:
	ld		a,3
	ld		($9000),a	; set address Y
	ld		a,b
	ld		($8000),a
	ld		a,0			; data mode
	ld		($9000),a
		
	ld		hl,320	; 320 pixels	
ClearScreen_xloop:	
	ld		($8000),a	; write black pixel
	scf					; clear carry
	ccf
	sbc		hl,de		; X loop
	jr		nz,ClearScreen_xloop
	
	ld		a,b		; if we cleared line 0, quit
	cp		0
	jr		z,ClearScreen_end
	sub		1		; otherwise, keep going
	ld		b,a
	jp		ClearScreen_yloop
ClearScreen_end:
	pop		de
	pop		hl
	pop		bc
	ret	


;******************************************************************************
; Sprite Data
;******************************************************************************

;
; Animations are hard-coded to be 4 frames. The tables consist of pointers to
; sprites and are cycled in sequence.
;

anim_idle:
	.dw	spr_garg + (0*23*31)
	.dw	spr_garg + (1*23*31)
	.dw	spr_garg + (2*23*31)
	.dw	spr_garg + (3*23*31)
	
anim_walk_left:
	.dw	spr_garg + (6*23*31)
	.dw	spr_garg + (7*23*31)
	.dw	spr_garg + (6*23*31)
	.dw	spr_garg + (7*23*31)

	
anim_walk_right:
	.dw	spr_garg + (4*23*31)
	.dw	spr_garg + (5*23*31)
	.dw	spr_garg + (4*23*31)
	.dw	spr_garg + (5*23*31)

;
; Blank sprite to erase with
;

spr_blank:
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	.db 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;
; The actual Garg sprites are appended as raw binary data to the ROM file
;
spr_garg:

.END
